@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2007, 2008,
-   2009, 2010 Free Software Foundation, Inc.
+   Copyright (C) 1999-2020 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -13,9 +12,8 @@
    Lesser General Public License for more details.
 
    You should have received a copy of the GNU Lesser General
-   Public License along with this library; if not, write to the
-   Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-   Boston, MA 02110-1301 USA */
+   Public License along with this library.  If not, see
+   <http://www.gnu.org/licenses/>. */
 
 /* First draft by Sergey Poznyakoff */
 
@@ -23,8 +21,6 @@
 # include <config.h>
 #endif
 
-#ifdef ENABLE_MH
-
 #include <sys/types.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -34,6 +30,7 @@
 #include <unistd.h>
 #include <errno.h>
 #include <dirent.h>
+#include <limits.h>
 
 #ifdef WITH_PTHREAD
 # ifdef HAVE_PTHREAD_H
@@ -54,19 +51,22 @@
 #include <mailutils/debug.h>
 #include <mailutils/envelope.h>
 #include <mailutils/error.h>
+#include <mailutils/errno.h>
 #include <mailutils/header.h>
 #include <mailutils/locker.h>
 #include <mailutils/message.h>
-#include <mailutils/mutil.h>
+#include <mailutils/util.h>
 #include <mailutils/property.h>
 #include <mailutils/stream.h>
 #include <mailutils/url.h>
 #include <mailutils/observer.h>
 #include <mailutils/io.h>
 #include <mailutils/cctype.h>
-#include <mailbox0.h>
-#include <registrar0.h>
-#include <amd.h>
+#include <mailutils/cstr.h>
+#include <mailutils/mh.h>
+#include <mailutils/sys/mailbox.h>
+#include <mailutils/sys/registrar.h>
+#include <mailutils/sys/amd.h>
 
 struct _mh_message
 {
@@ -97,7 +97,7 @@ _mh_next_seq (struct _amd_data *amd)
 /* Return current filename for the message.
    NOTE: Allocates memory. */
 static int
-_mh_cur_message_name (struct _amd_message *amsg, char **pname)
+_mh_cur_message_name (struct _amd_message *amsg, int abs, char **pname)
 {
   int status = 0;
   struct _mh_message *mhm = (struct _mh_message *) amsg;
@@ -108,12 +108,19 @@ _mh_cur_message_name (struct _amd_message *amsg, char **pname)
   status = mu_asprintf (&pnum, "%lu", (unsigned long) mhm->seq_number);
   if (status)
     return status;
-  len = strlen (amsg->amd->name) + 1 + strlen (pnum) + 1;
+  len = strlen (pnum) + 1;
+  if (abs)
+    len += strlen (amsg->amd->name) + 1;
   filename = malloc (len);
   if (filename)
     {
-      strcpy (filename, amsg->amd->name);
-      strcat (filename, "/");
+      if (abs)
+	{
+	  strcpy (filename, amsg->amd->name);
+	  strcat (filename, "/");
+	}
+      else
+	filename[0] = 0;
       strcat (filename, pnum);
       *pname = filename;
     }
@@ -182,6 +189,7 @@ mh_scan0 (mu_mailbox_t mailbox, size_t msgno MU_ARG_UNUSED, size_t *pcount,
   struct dirent *entry;
   int status = 0;
   struct stat st;
+  int need_sort = 0;
 
   if (amd == NULL)
     return EINVAL;
@@ -236,96 +244,141 @@ mh_scan0 (mu_mailbox_t mailbox, size_t msgno MU_ARG_UNUSED, size_t *pcount,
       if (!msg)
 	{
 	  msg = calloc (1, sizeof(*msg));
+	  status = _amd_message_append (amd, (struct _amd_message *) msg);
+	  if (status)
+	    {
+	      free (msg);
+	      break;
+	    }
 
 	  msg->seq_number = num;
 	  msg->amd_message.attr_flags = attr_flags;
-	  msg->amd_message.orig_flags = msg->amd_message.attr_flags;
 
-	  _amd_message_insert (amd, (struct _amd_message*) msg);
+	  need_sort = 1;
 	}
       else
 	{
 	  msg->amd_message.attr_flags = attr_flags;
-	  msg->amd_message.orig_flags = msg->amd_message.attr_flags;
 	}
     }
 
   closedir (dir);
 
-  if (do_notify)
+  if (need_sort)
+    amd_sort (amd);
+
+  if (status == 0)
     {
-      size_t i;
-
-      for (i = 0; i < amd->msg_count; i++)
+      if (do_notify)
 	{
-	  DISPATCH_ADD_MSG (mailbox, amd, i);
-	}
-    }
-  
-  if (stat (amd->name, &st) == 0)
-    amd->mtime = st.st_mtime;
+	  size_t i;
 
-  if (pcount)
-    *pcount = amd->msg_count;
-
-  /* Reset the uidvalidity.  */
-  if (amd->msg_count > 0)
-    {
-      if (amd->uidvalidity == 0)
-	{
-	  amd->uidvalidity = (unsigned long)time (NULL);
-	  /* Tell that we have been modified for expunging.  */
-	  if (amd->msg_count)
+	  for (i = 0; i < amd->msg_count; i++)
 	    {
-	      amd_message_stream_open (amd->msg_array[0]);
-	      amd_message_stream_close (amd->msg_array[0]);
-	      amd->msg_array[0]->attr_flags |= MU_ATTRIBUTE_MODIFIED;
+	      DISPATCH_ADD_MSG (mailbox, amd, i);
 	    }
 	}
-    }
+  
+      if (stat (amd->name, &st) == 0)
+	amd->mtime = st.st_mtime;
 
+      if (pcount)
+	*pcount = amd->msg_count;
+    }
   /* Clean up the things */
 
+  mu_locker_unlock (mailbox->locker);
   amd_cleanup (mailbox);
 #ifdef WITH_PTHREAD
   pthread_cleanup_pop (0);
 #endif
   return status;
 }
+
+static int
+mh_size_unlocked (struct _amd_data *amd, mu_off_t *psize)
+{
+  mu_off_t size = 0;
+  int rc;
+  struct stat st;
+  DIR *dir;
+  struct dirent *entry;
+  
+  dir = opendir (amd->name);
+  if (!dir)
+    return errno;
 
+  while ((entry = readdir (dir)))
+    {
+      if (*mu_str_skip_class (entry->d_name, MU_CTYPE_DIGIT) == 0)
+	{
+	  char *fname = mu_make_file_name (amd->name, entry->d_name);
+	  if (!fname)
+	    continue;
+	  if (stat (fname, &st))
+	    {
+	      rc = errno;
+	      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+			("can't stat %s: %s", fname, mu_strerror (rc)));
+	      free (fname);
+	      continue;
+	    }
+	  if (S_ISREG (st.st_mode))
+	    size += st.st_size;
+	}
+    }
+
+  *psize = size;
+  
+  closedir (dir);
+
+  return 0;
+}
+
+static int
+mh_size (mu_mailbox_t mailbox, mu_off_t *psize)
+{
+  struct _amd_data *amd = mailbox->data;
+  int rc;
+  
+  mu_monitor_wrlock (mailbox->monitor);
+#ifdef WITH_PTHREAD
+  pthread_cleanup_push (amd_cleanup, (void *)mailbox);
+#endif
+  mu_locker_lock (mailbox->locker);
+
+  rc = mh_size_unlocked (amd, psize);
+  
+  mu_locker_unlock (mailbox->locker);
+  mu_monitor_unlock (mailbox->monitor);
+#ifdef WITH_PTHREAD
+  pthread_cleanup_pop (0);
+#endif
+  return rc;
+}
+
+
 static int
 mh_qfetch (struct _amd_data *amd, mu_message_qid_t qid)
 {
-  char *p;
+  char *p = (char *)qid;
   size_t num = 0;
   int attr_flags = 0;
   struct _mh_message *msg;
 
-  p = qid + strlen (qid) - 1;
-  if (!mu_isdigit (*p))
-    return EINVAL;
-  
-  for (p--; p >= qid && mu_isdigit (*p); p--)
-    ;
-
-  if (p == qid)
-    return EINVAL;
-
-  num = strtoul (p + 1, NULL, 10);
-  
   if (*p == ',')
     {
       attr_flags |= MU_ATTRIBUTE_DELETED;
-      p--;
+      p++;
     }
 
-  if (*p != '/')
+  num = strtoul (p, &p, 10);
+  if ((num == ULONG_MAX && errno == ERANGE) || *p != 0)
     return EINVAL;
   
   msg = calloc (1, sizeof (*msg));
   msg->seq_number = num;
   msg->amd_message.attr_flags = attr_flags;
-  msg->amd_message.orig_flags = msg->amd_message.attr_flags;
   _amd_message_insert (amd, (struct _amd_message*) msg);
   return 0;
 }
@@ -352,6 +405,101 @@ _mh_msg_init (struct _amd_data *amd, struct _amd_message *amm)
   return 0;
 }
 
+static int
+_mh_msg_delete (struct _amd_data *amd, struct _amd_message *amm)
+{
+  int rc, status;
+  char *name;
+  char *argv[3];
+  /* FIXME: Cache this value */
+  const char *proc = mu_mhprop_get_value (mu_mh_profile, "rmmproc", NULL);
+  if (!proc)
+    return ENOSYS;
+
+  rc = amd->cur_msg_file_name (amm, 1, &name);
+  if (rc)
+    return rc;
+  
+  if (proc[0] == 0)
+    {
+      if (unlink (name))
+	rc = errno;
+    }
+  else
+    {
+      argv[0] = (char*) proc;
+      argv[1] = name;
+      argv[2] = NULL;
+      rc = mu_spawnvp (proc, argv, &status);
+    }
+  free (name);
+  return rc;
+}
+
+
+static int
+mh_remove (struct _amd_data *amd)
+{
+  return amd_remove_dir (amd->name);
+}
+
+
+static int
+mh_get_property (mu_mailbox_t mailbox, mu_property_t *pprop)
+{
+  struct _amd_data *amd = mailbox->data;
+  mu_property_t property = NULL;
+  struct mu_mh_prop *mhprop;
+  const char *p;
+  
+  mhprop = calloc (1, sizeof (mhprop[0]));
+  if (!mhprop)
+    return ENOMEM;
+  p = mu_mhprop_get_value (mu_mh_profile, "mh-sequences",
+			   MU_MH_SEQUENCES_FILE);
+  mhprop->filename = mu_make_file_name (amd->name, p);
+  mu_property_create_init (&property, mu_mh_property_init, mhprop);
+  mu_mailbox_set_property (mailbox, property);
+      
+  /*FIXME mu_property_set_value (property, "TYPE", "MH", 1);*/
+  *pprop = property;
+  return 0;
+}
+
+
+
+static int
+mh_translate (mu_mailbox_t mbox, int cmd, size_t from, size_t *to)
+{
+  struct _amd_data *amd = mbox->data;
+  struct _mh_message msg, *mp;
+  size_t n;
+
+  /* Make sure the mailbox has been scanned */ 
+  mu_mailbox_messages_count (mbox, &n);
+    
+  switch (cmd)
+    {
+    case MU_MAILBOX_UID_TO_MSGNO:
+      msg.seq_number = from;
+      if (amd_msg_lookup (amd, (struct _amd_message*) &msg, &n))
+	return MU_ERR_NOENT;
+      *to = n;
+      break;
+
+    case MU_MAILBOX_MSGNO_TO_UID:
+      mp = (struct _mh_message *) _amd_get_message (amd, from);
+      if (!mp)
+	return MU_ERR_NOENT;
+      *to = mp->seq_number;
+      break;
+
+    default:
+      return ENOSYS;
+    }
+  return 0;
+}
+
 
 
 int
@@ -375,15 +523,17 @@ _mailbox_mh_init (mu_mailbox_t mailbox)
   amd->msg_cmp = mh_message_cmp;
   amd->message_uid = mh_message_uid;
   amd->next_uid = _mh_next_seq;
+  amd->remove = mh_remove;
+  amd->mailbox_size = mh_size;
   
-  /* Set our properties.  */
-  {
-    mu_property_t property = NULL;
-    mu_mailbox_get_property (mailbox, &property);
-    mu_property_set_value (property, "TYPE", "MH", 1);
-  }
+  mailbox->_get_property = mh_get_property;
+  mailbox->_translate = mh_translate;
 
+  if (mu_mhprop_get_value (mu_mh_profile, "rmmproc", NULL))
+    amd->delete_msg = _mh_msg_delete;
+    
+  
   return 0;
 }
 
-#endif
+
